# General concepts

## Preview
Throughout calng, a distinction is made between preview data and "full" data.
This distinction has two reasons behind it:

1. the data we send should be formatted nicely for GUI preview
    - some Karabo-specifics (`ImageData` element wrapping)
    - there should be no `nan` or `inf` values
    - optionally: downsampling for bandwidth reasons
2. we don't want to send full detector data to the guiserver (this would simply make the GUI unusable)
    - instead, we reduce data down to one frame per train
    - additionally, the guiserver will throttle to 2 Hz by default


This is all handled in two different steps and may need to be configured in two different places (until we figure out a nice way to package the configuration).
Step 1 happens on the device which produces the preview output - can be [correction device](devices/correction-devices.md#single-module-preview) or [preview assembler](devices/matchers.md#preview-assemblers).
Step 2 happens as early as possible - inside the correction devices.

For step 1, the following configurations are available:

- [NaN replacement](schemas/DetectorAssembler.md#preview.assembled.replaceNanWith): Karabo widgets don't play well with `nan` values.
  Bad pixel masking by default sets bad pixels to `nan` and the space between modules when assembled is by default filled with `nan`.
  Therefore, it is up to the operator to decide which value to show for these pixels.
- Downsampling [factor](schemas/DetectorAssembler.md#preview.assembled.downsamplingFactor) and [function](schemas/DetectorAssembler.md#preview.assembled.downsamplingFunction): allows downsampling of preview image resolution; mostly relevant over slow connections, hopefully not in control room.
  Downsampling (enabled if the factor is greater than one) is recursive halving, so factors can only be powers of two and image dimensions must be multiples of the factor.
  The downsampling function defines the downscaling kernel.
  This function is applied to each 2x2 group of pixels during recursive halving.

For step 2, the options available (in the case of fast detectors) in terms of frame reduction are:

- How to select which frame to preview (for non-negative indices)
	- Frame can be extracted directly from image ndarray (`frame` mode), by looking up corresponding frame in cell table (`cell` mode), or by looking up in pulse table (`pulse` mode)
    - For XTDF detectors, the mapping tables are typically `image.cellId` and `image.pulseId`.
	  Which selection mode makes sense depends on detector, veto pattern, and experimental setup.
    - If the specified cell or pulse is not found in the respective table, a warning is issued and the first frame is sent instead.
- Which frame or statistic to send
    - If the index is non-negative, a single frame is sliced from the image data.
	  How this frame is found depends on the the index selection mode (previous point).
    - If the index is negative, a statistic is computed across all frames for a summary preview.
      Note that this is done individually per pixel and that `NaN` values are ignored.
      Options are:
        - `-1` for max
        - `-2` for mean
        - `-3` for sum
        - `-4` for standard deviation

See also:

- [single module preview](devices/correction-devices.md#single-module-preview)
- [assembled preview](devices/middlelayer.md#assembled-preview-configuration)
- [preview assemblers](devices/matchers.md#preview-assemblers)

## Shared memory handles
To avoid expensive memory copies, correction devices (depending on [configuration](schemas/BaseCorrection.md#useShmemHandles)) put corrected big image data into shared memory.
This means that they can send results to group matchers or other devices on the same node very cheaply.
It also means that something like a group matcher is *necessary* for anything *not* on the same node to get access to the data; the group matcher will "dereference" the "shared memory handles" sent by the correction devices.

The following is a description of the protocol used for this purpose.
It is a bit technical and mostly relevant in case you want to develop a device to run on a correction node, exploiting the shared memory handles on there for cheap data access.

`calng` uses shared memory to pass detector data around between devices running on the same node, in particular the actual correction device, train matchers, bridges and other analysis devices such as `MetroProcessor`. This avoids the serialization and memory transfer cost that even on the same node can be substantial at rates exceeding tens of Gb/s.

The hash being sent from the `dataOutput` pipeline of a correction contains the `VECTOR_STRING` key `calngShmemPaths`. This may list a number of `NDARRAY` paths in same hash that do not actually contain data, but instead contain point to a shared memory region with the actual data. The format of this string is:

```
{name}${dtype}${shape}${index}
```

* `name` a unique identifier for the shared memory region, usually containing the correction device ID and  pipeline name
* `dtype` the string representation of the numpy data type, i.e. recoverable via `np.dtype(dtype)`
* `shape` a comma-separated list of axis lengths
* `index` an integer pointing to the position along the first array axis acting as a ring buffer

e.g. `SQS_NQS_DSSC/CAL/CORRECT01_Q1M2:dataOutput$float16$50,512,128,60$5`

The shared memory is allocated via the [POSIX shared memory API](https://man7.org/linux/man-pages/man7/shm_overview.7.html) - as of Python 3.8, `multiprocessing.shared_memory` provides classes wrapping the POSIX API.
For convenience, [`shmem_utils.py`](https://git.xfel.eu/calibration/calngutils/-/blob/master/src/calngUtils/shmem_utils.py) in `calngUtils` provides some wrappers for sending and receiving using our protocol.
The `ShmemCircularBufferReceiver` in there handles the following:

* The allocated shared memory region (indicated by `name`) may change on any train, but can be assumed to be fairly constant and its mapping can be cached.
  If and only if `name` stays the same, it can be assumed to be the same shared memory region of the same size.
* Even with an unchanged memory region name, the `dtype` and `shape` may also change on any train and require wrapping a new `ndarray` around it to correctly find the data.
* The array described by `dtype` and `shape` is used as a circular buffer with `0 ≤ index < shape[0]`.
  It always starts at no offset, but may be smaller than the entire shared memory region.
  There is no guarantee how long any given entry stays around, but is currently ≥ 10 s.
* The lifetime of the shared memory region is bound to its creator (a correction device) and it should be assumed to be unlinked once it shuts down or `name` changed.
  POSIX semantics guarantee any consumer the memory stays valid until closed on their side.
