import operator

from karabo.bound import (
    DOUBLE_ELEMENT,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
)
import numpy as np
from .base_kernel import BaseArbiterKernel


class ReduceAndThreshold(BaseArbiterKernel):
    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            OVERWRITE_ELEMENT(schema)
            .key(prefix)
            .setNewDescription(
                "This kernel allows custom frame masks based on comparing some "
                "reduction along modules over each frame to some threshold. "
                "Integrated intensity will be used as a running example in the "
                "descriptions of the properties under this node; these settings (used "
                "with the example integrated intensity addon) would yield a frame "
                "mask of frames with total (across modules) integrated intensity of "
                "at least [threshold] ADUs."
            )
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.key")
            .description(
                "The key in the input hashes which holds the data values for reduction "
                "and comparison. Typically some summary computed per frame by an "
                "addon; for example, 'image.integratedIntensity'"
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{prefix}.threshold")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.reduction")
            .description(
                "The reduction to apply across modules before comparing with the "
                "threshold. For example 'sum'."
            )
            .assignmentOptional()
            .defaultValue("sum")
            .options("sum,mean,max,min,std")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.comparator")
            .description(
                "The comparator to use for comparing the reduced per-frame values to "
                "the threshold. For the integrated intensity example, 'ge'. The "
                "reduced values will be on the left and the threshold will be on the "
                "right, so the example would be compute: "
                "mask = sum(integrated intensity) ≥ threshold."
            )
            .assignmentOptional()
            .defaultValue("ge")
            .options("lt,le,gt,ge,eq,ne")
            .reconfigurable()
            .commit(),
        )

    def reconfigure(self, config):
        if config.has("key"):
            self._key = config.get("key")
        if config.has("threshold"):
            self._threshold = config.get("threshold")
        if config.has("reduction"):
            self._reduction = getattr(np, config.get("reduction"))
        if config.has("comparator"):
            self._comparator = getattr(operator, config.get("comparator"))

    def consider(self, train_id, sources, num_frames, mask, out_hash):
        res = np.zeros(num_frames, dtype=bool)
        mask = mask.astype(bool, copy=False)
        sources_with_key = [
            data[self._key][mask]
            for (data, _) in sources.values()
            if data.has(self._key)
        ]
        if not sources_with_key:
            raise KeyError(f"No sources had '{self._key}'")
        res[mask] = self._comparator(
            self._reduction(sources_with_key, axis=0), self._threshold
        )
        return res
