import concurrent.futures
import functools
import itertools
import math
import threading
from importlib.metadata import entry_points
from timeit import default_timer

import dateutil.parser
import numpy as np
from calngUtils import (
    device as device_utils,
    geom_utils,
    scene_utils,
    shmem_utils,
    timing,
    trackers,
)
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    INPUT_CHANNEL,
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    UINT64_ELEMENT,
    VECTOR_STRING_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    ChannelMetaData,
    Hash,
    MetricPrefix,
    PythonDevice,
    Schema,
    State,
    Timestamp,
    Unit,
)
from karabo.common.api import KARABO_SCHEMA_DISPLAY_TYPE_SCENES as DT_SCENES

from . import scenes
from .preview_utils import PreviewFriend
from .utils import StateContext, WarningContextSystem, WarningLampType, subset_of_hash
from ._version import version as deviceVersion

PROCESSING_STATE_TIMEOUT = 10


class TrainFromTheFutureException(BaseException):
    pass


@device_utils.with_config_overlay
@device_utils.with_unsafe_get
@KARABO_CLASSINFO("BaseCorrection", deviceVersion)
class BaseCorrection(PythonDevice):
    _available_addons = []  # classes, filled by add_addon_nodes using entry_points
    _base_output_schema = None  # subclass must set constructor
    _constant_enum_class = None  # subclass must set
    _correction_steps = None  # subclass must set
    _kernel_runner_class = None  # subclass must set (ex.: dssc_gpu.DsscGpuRunner)
    _image_data_path = "image.data"  # customize for *some* subclasses
    _cell_table_path = "image.cellId"
    _pulse_table_path = "image.pulseId"
    _warn_memory_cell_range = True  # can be disabled for some detectors
    _cuda_pin_buffers = False
    _use_shmem_handles = True

    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("state")
            .setNewDefaultValue(State.INIT)
            .commit(),

            INPUT_CHANNEL(expected)
            .key("dataInput")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataInput.onSlowness")
            .setNewTags("managed")
            .setNewDefaultValue("queueDrop")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataInput.maxQueueLength")
            .setNewTags("managed")
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("fastSources")
            .displayedName("Fast data sources")
            .description(
                "Sources to get data from. Only incoming hashes from these sources "
                "will be processed. This will typically be a single entry of the form: "
                "'[instrument]_DET_[detector]/DET/[channel]:xtdf'."
            )
            .assignmentOptional()
            .defaultValue([])
            .commit(),

            UINT32_ELEMENT(expected)
            .key("outputShmemBufferSize")
            .tags("managed")
            .displayedName("Output buffer size limit")
            .unit(Unit.BYTE)
            .metricPrefix(MetricPrefix.GIGA)
            .description(
                "Corrected trains are written to shared memory locations. These are "
                "pre-allocated and re-used (circular buffer). This parameter "
                "determines how much memory to set aside for that buffer."
            )
            .assignmentOptional()
            .defaultValue(10)
            .commit(),

            VECTOR_STRING_ELEMENT(expected)
            .key("availableScenes")
            .setSpecialDisplayType(DT_SCENES)
            .readOnly()
            .initialValue(["overview", "constant_overrides"])
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useShmemHandles")
            .tags("managed")
            .displayedName("Use shared memory")
            .description(
                "If enabled, shared memory handles will be used to avoid copying "
                "the main image data on the dataOutput channel.  Note that these "
                "handles need to be dereferenced (typically by a group matcher) to "
                "get the data off the correction node."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useInfiniband")
            .tags("managed")
            .displayedName("Use infiniband")
            .description(
                "If enabled, device will during initialization try to bind its main "
                "data output channel (dataOutput) to its node's infiniband interface. "
                "Default interface is used if no 'ib0' interface is found."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .assignmentOptional()
            .defaultValue("")
            .commit(),
        )

        (
            NODE_ELEMENT(expected)
            .key("dataFormat")
            .displayedName("Data format (in/out)")
            .commit(),

            STRING_ELEMENT(expected)
            .key("dataFormat.outputImageDtype")
            .tags("managed")
            .displayedName("Output image data dtype")
            .description(
                "The (numpy) dtype to use for outgoing image data. Input is cast to "
                "float32, corrections are applied, and only then will the result be "
                "cast to outputImageDtype. Be aware that casting to integer type "
                "causes truncation rather than rounding."
            )
            # TODO: consider adding rounding / binning for integer output
            .options("float32,float16,uint16")
            .assignmentOptional()
            .defaultValue("float32")
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(expected)
            .key("dataFormat.inputFrames")
            .displayedName("Frames")
            .description("Number of frames on input")
            .readOnly()
            .initialValue(0)
            .commit(),

            STRING_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .tags("managed")
            .displayedName("Output axis order")
            .description(
                "Axes of main data output can be reordered after correction. Axis "
                "order is specified as string consisting of 'f' (frames), 'ss' (slow "
                "scan), and 'ff' (fast scan). Anything but the default f-ss-fs order "
                "implies reordering - this axis ordering is based on the data we get "
                "from the detector (/ receiver), so how ss and fs maps to what may be "
                "considered x and y is detector-dependent."
            )
            .options("f-ss-fs,f-fs-ss,ss-f-fs,ss-fs-f,fs-f-ss,fs-ss-f")
            .assignmentOptional()
            .defaultValue("f-ss-fs")
            .reconfigurable()
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.inputDataShape")
            .displayedName("Input data shape")
            .description(
                "Image data shape in incoming data (from reader / DAQ). Updated based "
                "on latest train processed. Note that axis order may look wonky due to "
                "DAQ quirk."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.outputDataShape")
            .displayedName("Output data shape")
            .description(
                "Image data shape for data output from this device. Takes into account "
                "axis reordering, if applicable. Even without that, "
                "workarounds.overrideInputAxisOrder likely makes this differ from "
                "dataFormat.inputDataShape."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.cellId")
            .readOnly()
            .initialValue([])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected)
            .key("dataFormat.pulseId")
            .readOnly()
            .initialValue([])
            .commit(),

            NODE_ELEMENT(expected)
            .key("workarounds")
            .displayedName("Workarounds")
            .commit(),

            BOOL_ELEMENT(expected)
            .key("workarounds.overrideInputAxisOrder")
            .tags("managed")
            .displayedName("Override input axis order")
            .description(
                "The shape of the image data ndarray as received from the "
                "DataAggregator is sometimes wrong - the axes are actually in a "
                "different order than the ndarray shape suggests. If this flag is on, "
                "the shape of the ndarray will be overridden with the axis order which "
                "was expected."
            )
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            UINT64_ELEMENT(expected)
            .key("workarounds.trainFromFutureThreshold")
            .tags("managed")
            .displayedName("Spurious future train ID threshold")
            .description(
                "Some detectors occasionally send a train with incorrect and much too "
                "large train ID. To avoid these 'future trains' from interfering with "
                "train matching, use this threshold to discard them immediately. If a "
                "train arrives with an ID which exceeds the current train ID from the "
                "time server by more than this threshold, the train is ignored."
            )
            .assignmentOptional()
            .defaultValue(10000)
            .reconfigurable()
            .commit(),
        )

        (
            SLOT_ELEMENT(expected)
            .key("loadMostRecentConstants")
            .tags("managed")
            .displayedName("Load most recent constants")
            .description(
                "Calling this slot will flush all constant buffers and cause the "
                "device to start querying CalCat for the most recent constants (all "
                "constants applicable for this device) available with the currently "
                "set constant parameters. This should typically be called after "
                "instantiating pipeline, after changing parameters, or after "
                "generating new constants."
            )
            .commit()
        )

        (
            STRING_ELEMENT(expected)
            .key("inputDataState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),

            STRING_ELEMENT(expected)
            .key("deviceInternalsState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),

            STRING_ELEMENT(expected)
            .key("processingState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("NORMAL")
            .commit(),
        )

        # just measurements and counters to display
        (
            UINT64_ELEMENT(expected)
            .key("trainId")
            .displayedName("Train ID")
            .description("ID of latest train processed by this device.")
            .readOnly()
            .initialValue(0)
            .commit(),

            NODE_ELEMENT(expected)
            .key("performance")
            .displayedName("Performance measures")
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(100)
            .info("Processing too slow to reach 10 Hz")
            .needsAcknowledging(False)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.rate")
            .displayedName("Rate")
            .description(
                "Actual rate with which this device gets, processes, and sends trains. "
                "This is a simple windowed moving average."
            )
            .unit(Unit.HERTZ)
            .readOnly()
            .initialValue(0)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.inputDelay")
            .displayedName("Input delay")
            .description(
                "Measured delay on the input channel. Measured by difference between "
                "'now' and timestamp in input metadata."
            )
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("performance.ratioOfRecentTrainsReceived")
            .description(
                "Of the latest trains (from last received train, going back "
                "[some buffer range]), how many did we receive? This estimate is "
                "updated when new trains come in, so is unreliable if nothing is "
                "coming at all."
            )
            .unit(Unit.PERCENT)
            .readOnly()
            .initialValue(0)
            .commit(),
        )

        # this node will be filled out by subclass
        (
            NODE_ELEMENT(expected)
            .key("corrections")
            .displayedName("Correction steps")
            .commit(),
        )
        (
            VECTOR_STRING_ELEMENT(expected)
            .key("warningLamps")
            .assignmentOptional()
            .defaultValue(
                [
                    "deviceInternalsState",
                    "inputDataState",
                    "processingState",
                ]
            )
            .commit()
        )
        # subclasses can populate with parameters / controls for plugins
        (
            NODE_ELEMENT(expected)
            .key("addons")
            .displayedName("Addons")
            .commit(),
        )

    def __init__(self, config):
        if config.get("useInfiniband", default=True):
            from PipeToZeroMQ.utils import find_infiniband_ip

            ib_ip = find_infiniband_ip()
            config["dataOutput.hostname"] = ib_ip
        super().__init__(config)

        self.sources = set(config.get("fastSources"))

        self._buffer_lock = threading.Lock()
        self._last_processing_started = 0  # used for processing time and timeout
        self._last_train_id_processed = 0  # used to keep track (and as fallback)
        self._last_cell_table = None  # used to check whether to update property
        self._last_pulse_table = None  # ditto - though not all detectors have this

        def constant_override_fun(friend_fun, constant, preserve_fields):
            def aux():
                self.flush_constants(
                    constants={constant}, preserve_fields=preserve_fields
                )
                with self.warning_context(
                    f"foundConstants.{constant.name}.state"
                ) as warn:
                    try:
                        constant_data = getattr(self.calcat_friend, friend_fun)(
                            constant
                        )
                        self.kernel_runner.load_constant(constant, constant_data)
                    except Exception as e:
                        warn(f"Failed to load {constant}: {e}")
                # note: consider if multi-constant buffers need special care
            self._lock_and_update(aux)

        for constant, (friend_fun, preserved_fields, slot_name) in itertools.product(
            self._constant_enum_class,
            (
                ("get_constant_version", set(), "loadMostRecent"),
                (
                    "get_constant_from_constant_version_id",
                    {"constantVersionId"},
                    "overrideConstantFromVersion",
                ),
                (
                    "get_constant_from_file",
                    {"dataFilePath", "dataSetName"},
                    "overrideConstantFromFile",
                ),
            ),
        ):
            self.KARABO_SLOT(
                functools.partial(
                    constant_override_fun,
                    friend_fun,
                    constant,
                    preserved_fields,
                ),
                slotName=f"foundConstants_{constant.name}_{slot_name}",
                numArgs=0,
            )

        self.KARABO_SLOT(self.loadMostRecentConstants)
        self.KARABO_SLOT(self.requestScene)

        self.registerInitialFunction(self._initialization)

    def _initialization(self):
        self.updateState(State.INIT)
        self.warning_context = WarningContextSystem(
            self,
            on_success={
                f"foundConstants.{constant.name}.state": "ON"
                for constant in self._constant_enum_class
            },
        )
        self.log.DEBUG("Opening shmem buffer")
        self._shmem_buffer = shmem_utils.ShmemCircularBuffer(
            self.get("outputShmemBufferSize") * 2**30,
            # TODO: have it just allocate memory, then set_shape later per train
            (1,),
            np.float32,
            self.getInstanceId() + ":dataOutput",
        )
        if self._cuda_pin_buffers:
            self.log.DEBUG("Trying to pin the shmem buffer memory")
            self._shmem_buffer.cuda_pin()
        self._shmem_receiver = shmem_utils.ShmemCircularBufferReceiver()

        # CalCat friend comes before kernel runner
        with self.warning_context(
            "deviceInternalsState", WarningLampType.CALCAT_CONNECTION
        ) as warn:
            try:
                self.calcat_friend = self._calcat_friend_class(self)
            except Exception as e:
                warn(f"Failed to connect to CalCat: {e}")
                # TODO: add raw fallback mode if CalCat fails (raw data still useful)
                return
        self.kernel_runner = self._kernel_runner_class(self)
        self.preview_friend = PreviewFriend(self, self._preview_outputs)
        # TODO: can be static OVERWRITE_ELEMENT in expectedParameters
        self["availableScenes"] = self["availableScenes"] + [
            f"preview:{spec.name}" for spec in self._preview_outputs
        ]

        self._buffered_status_update = Hash()
        self._processing_time_tracker = trackers.ExponentialMovingAverage(
            alpha=0.3
        )
        self._rate_tracker = trackers.WindowRateTracker()
        self._input_delay_tracker = trackers.ExponentialMovingAverage(
            alpha=0.3
        )
        self._rate_update_timer = timing.RepeatingTimer(
            interval=1,
            callback=self._update_rate_and_state,
        )
        self._train_ratio_tracker = trackers.TrainRatioTracker()

        self.KARABO_ON_DATA("dataInput", self.input_handler)
        self.KARABO_ON_EOS("dataInput", self.handle_eos)

        self._enabled_addons = []
        for addon_class in self._available_addons:
            addon_prefix = f"addons.{addon_class.__name__}"
            if not self.get(f"{addon_prefix}.enable"):
                continue
            addon = addon_class(self._parameters[addon_prefix])
            addon._device = self
            addon._prefix = addon_prefix
            self._enabled_addons.append(addon)
        if (
            (self.get("useShmemHandles") != self._use_shmem_handles)
            or self._enabled_addons
        ):
            schema_override = Schema()
            output_schema_override = self.__class__._base_output_schema(
                use_shmem_handles=self.get("useShmemHandles")
            )
            for addon in self._enabled_addons:
                addon.extend_output_schema(output_schema_override)
            (
                # add channel to schema override to re-inject with new output schema
                OUTPUT_CHANNEL(schema_override)
                .key("dataOutput")
                .dataSchema(output_schema_override)
                .commit(),
            )
            # update device schema in one fell swoop
            self.updateSchema(schema_override)

        self._geometry = None
        def _on_geometry(geometry):
            self._geometry = geometry
            for addon in self._enabled_addons:
                addon.on_new_geometry(self._geometry)
        geom_utils.subscribe_to_geometry_bound(
            self,
            self.get("geometryDevice"),
            on_geometry=_on_geometry,
            warn_fun=self.log_status_warn,
        )

        self.updateState(State.ON)

    def __del__(self):
        del self.shmem_buffer
        super().__del__()

    def preReconfigure(self, config):
        # validation
        for ts_path in (
            "constantParameters.deviceMappingSnapshotAt",
            "constantParameters.constantVersionEventAt",
        ):
            if config.has(ts_path):
                ts_string = config.get(ts_path)
                if ts_string.strip() == "":
                    config.set(ts_path, "")
                else:
                    # allow exception to reach operator
                    timestamp = dateutil.parser.isoparse(ts_string)
                    config.set(ts_path, timestamp.isoformat())

        with self.config_overlay(config):
            if config.has("constantParameters.deviceMappingSnapshotAt"):
                self.calcat_friend.flush_pdu_mapping()

            if (
                runner_update := subset_of_hash(
                    config, "corrections", "dataFormat", "constantParameters"
                )
            ):
                self.kernel_runner.reconfigure(runner_update)

            if config.has("preview"):
                self.preview_friend.reconfigure(config["preview"])

            if config.has("addons"):
                # note: can avoid iterating, but it's not that costly
                for addon in self._enabled_addons:
                    full_path = f"addons.{addon.__class__.__name__}"
                    if config.has(full_path):
                        addon.reconfigure(config[full_path])

    def _lock_and_update(self, method, background=True):
        # TODO: securely handle errors (postReconfigure may succeed, device state not)
        def runner():
            with self._buffer_lock, StateContext(self, State.CHANGING):
                method()

        if background:
            threading.Thread(target=runner, daemon=True).start()
        else:
            runner()

    def loadMostRecentConstants(self):
        def aux():
            self.flush_constants()
            # TODO: ignore irrelevant constants (like agpid thresholding in fixed gain)
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future_to_constant = {
                    executor.submit(
                        self.calcat_friend.get_constant_version,
                        constant,
                    ): constant
                    for constant in self._constant_enum_class
                }
                for future in concurrent.futures.as_completed(future_to_constant):
                    constant = future_to_constant[future]
                    with self.warning_context(
                        f"foundConstants.{constant.name}.state"
                    ) as warn:
                        try:
                            constant_data = future.result()
                            self.kernel_runner.load_constant(constant, constant_data)
                        except Exception as e:
                            warn(f"Failed to load {constant}: {e}")
        self._lock_and_update(aux)

    def flush_constants(self, constants=None, preserve_fields=None):
        """Reset constant buffers; the correction maps (often buffers on GPU) should be
        set to identity values and constant caches cleared. By default do this for all
        constants - a subset can instead be specified via the constants parameter
        (should be iterable of constant enum values).

        After flushing, the kernel runner will be instructed to reload any remaining
        cached constants after flushing. This can take a second.

        The parameter preserve_fields is passed to the CalCat friend; this is used for
        overrides; in those cases, we want to get value from an overridden field after
        flushing everything else, so don't wipe it out.
        """
        if constants is None:
            constants = set(self._constant_enum_class)
        else:
            constants = set(constants)
        self.kernel_runner.flush_buffers(constants)
        for field_name, _, used_constants in self._correction_steps:
            if (
                used_constants
                and None not in used_constants
                and constants >= used_constants
            ):
                self.set(f"corrections.{field_name}.available", False)
        self.calcat_friend.flush_constants(constants, preserve_fields)
        self._reload_constants_from_cache_to_runner(constants)

    def log_status_info(self, msg):
        self.log.INFO(msg)
        self.set("status", msg)

    def log_status_warn(self, msg):
        self.log.WARN(msg)
        self.set("status", msg)

    def requestScene(self, params):
        payload = Hash()
        name = params.get("name", default="")
        payload["success"] = True
        if name == "constant_overrides":
            payload["data"] = scenes.correction_device_constant_overrides(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
            )
        elif name.startswith("preview:"):
            channel_name = name[len("preview:") :]
            payload["data"] = scenes.correction_device_preview(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
                name=channel_name,
            )
        elif name.startswith("browse_schema"):
            if ":" in name:
                prefix = name[len("browse_schema:") :]
            else:
                prefix = "managed"
            payload["data"] = scene_utils.recursive_subschema_scene(
                self.getInstanceId(),
                self.getFullSchema(),
                prefix,
            )
        else:
            # overview by default
            name = "overview"
            payload["data"] = scenes.correction_device_overview(
                device_id=self.getInstanceId(),
                schema=self.getFullSchema(),
            )
        payload["name"] = name
        response = Hash()
        response["type"] = "deviceScene"
        response["origin"] = self.getInstanceId()
        response["payload"] = payload
        self.reply(response)

    def _get_data_from_hash(self, data_hash):
        """Will get image data, cell table, pulse table, and list of other arrays from
        the data hash. Assumes XTDF (image.data, image.cellId, image.pulseId, and no
        other data), non-XTDF subclass can override. Cell and pulse table can be
        None."""

        image_data = data_hash.get(self._image_data_path)
        cell_table = data_hash[self._cell_table_path].ravel()
        pulse_table = data_hash[self._pulse_table_path].ravel()
        num_frames = cell_table.size
        # DataAggregator typically tells us the wrong axis order
        if self.unsafe_get("workarounds.overrideInputAxisOrder"):
            # TODO: check that frames are always axis 0
            expected_shape = self.kernel_runner.expected_input_shape(
                num_frames
            )
            if expected_shape != image_data.shape:
                image_data.shape = expected_shape
        return (
            num_frames,
            image_data,  # not ndarray (runner can do that)
            cell_table,
            pulse_table,
        )

    def _write_output(self, data, old_metadata):
        """For dataOutput: reusing incoming data hash and setting source and timestamp
        to be same as input"""
        metadata = ChannelMetaData(
            old_metadata.get("source"),
            Timestamp.fromHashAttributes(old_metadata.getAttributes("timestamp")),
        )

        channel = self.signalSlotable.getOutputChannel("dataOutput")
        channel.write(data, metadata, copyAllData=False)
        channel.update(safeNDArray=True)

    def _reload_constants_from_cache_to_runner(self, constants=None):
        # TODO: gracefully handle change in constantParameters.memoryCells
        if constants is None:
            constants = set(self._constant_enum_class)
        else:
            constants = set(constants)

        with self.calcat_friend.cached_constants_lock:
            for (
                constant,
                data,
            ) in self.calcat_friend.cached_constants.items():
                if constant not in constants:
                    continue
                with self.warning_context(
                    f"foundConstants.{constant.name}.state"
                ) as warn:
                    try:
                        self.log_status_info(f"Reload constant {constant.name}")
                        self.kernel_runner.load_constant(constant, data)
                    except Exception as e:
                        warn(f"Failed to reload {constant.name}: {e}")

    def input_handler(self, data_hash, metadata):
        """Main handler for data input: Do a few simple checks to determine whether to
        even try processing. If yes, will pass data and information to process_data
        method provided by subclass."""

        # Is device even ready for this?
        state = State[self.unsafe_get("state")]
        if state is State.ERROR:
            # in this case, we should have already issued warning
            return

        with self.warning_context(
            "deviceInternalsState", WarningLampType.CORRECTION_RUNNER
        ) as warn:
            if self.kernel_runner is None or state is State.INIT:
                warn("Received data before correction device was ready")
                return

        with self.warning_context(
            "inputDataState", WarningLampType.MISC_INPUT_DATA
        ) as warn:
            source = metadata.get("source")

            if source not in self.sources:
                return
            elif not data_hash.has(self._image_data_path):
                warn("Ignoring hash without image node")
                return

        with self.warning_context(
            "inputDataState",
            WarningLampType.EMPTY_HASH,
        ) as warn:
            self._shmem_receiver.dereference_shmem_handles(data_hash)
            try:
                (
                    num_frames,
                    image_data,
                    cell_table,
                    pulse_table,
                    *additional_data
                ) = self._get_data_from_hash(data_hash)
            except RuntimeError as err:
                warn(
                    "Failed to load data from hash, probably empy hash from DAQ. "
                    f"(Is detector sending data?)\n{err}"
                )
                self._maybe_update_cell_and_pulse_tables(None, None)
                return

        # no more common reasons to skip input, so go to processing
        self._last_processing_started = default_timer()
        if state is State.ON:
            self.updateState(State.PROCESSING)
            self.log_status_info("Processing data")

        self._maybe_update_cell_and_pulse_tables(cell_table, pulse_table)
        timestamp = Timestamp.fromHashAttributes(
            metadata.getAttributes("timestamp")
        )

        self._check_train_id_and_time(timestamp)

        if self._warn_memory_cell_range:
            with self.warning_context(
                "processingState", WarningLampType.MEMORY_CELL_RANGE
            ) as warn:
                if (
                    self.unsafe_get("constantParameters.memoryCells")
                    <= cell_table.max()
                ):
                    warn("Input cell IDs out of range of constants")

        # TODO: change ShmemCircularBuffer API so we don't have to poke like this
        if (
            (output_shape := self.kernel_runner.expected_output_shape(num_frames))
            != self._shmem_buffer._buffer_ary.shape[1:]
            or self.kernel_runner._output_dtype != self._shmem_buffer._buffer_ary.dtype
        ):
            self.set("dataFormat.outputDataShape", list(output_shape))
            self.log.DEBUG("Updating shmem buffer shape / dtype")
            self._shmem_buffer.change_shape(
                output_shape, self.kernel_runner._output_dtype
            )
        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        with self.warning_context(
            "processingState", WarningLampType.CORRECTION_RUNNER
        ):
            corrections, processed_buffer, previews = self.kernel_runner.correct(
                image_data, cell_table, *additional_data
            )
            data_hash["corrections"] = corrections
            for addon in self._enabled_addons:
                addon.post_correction(
                    timestamp.getTrainId(),
                    processed_buffer,
                    cell_table,
                    pulse_table,
                    data_hash,
                )
            self.kernel_runner.reshape(processed_buffer, out=buffer_array)

        with self.warning_context(
            "processingState", WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            self.preview_friend.write_outputs(
                *previews,
                timestamp=timestamp,
                cell_table=cell_table,
                pulse_table=pulse_table,
                warn_fun=warn,
            )

        for addon in self._enabled_addons:
            addon.post_reshape(
                timestamp.getTrainId(), buffer_array, cell_table, pulse_table, data_hash
            )

        if self.unsafe_get("useShmemHandles"):
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        self._write_output(data_hash, metadata)
        self._processing_time_tracker.update(
            default_timer() - self._last_processing_started
        )
        self._buffered_status_update.set(
            "performance.processingTime", self._processing_time_tracker.get() * 1000
        )
        self._rate_tracker.update()

    def _please_send_me_cached_constants(self, constants, callback):
        for constant in constants:
            if constant in self.calcat_friend.cached_constants:
                callback(constant, self.calcat_friend.cached_constants[constant])

    def _check_train_id_and_time(self, timestamp):
        train_id = timestamp.getTrainId()
        # check time server connection
        with self.warning_context(
            "deviceInternalsState", WarningLampType.TIMESERVER_CONNECTION
        ) as warn:
            my_timestamp = self.getActualTimestamp()
            my_train_id = my_timestamp.getTrainId()
            self._input_delay_tracker.update(
                (my_timestamp.toTimestamp() - timestamp.toTimestamp()) * 1000
            )
            self._buffered_status_update.set(
                "performance.inputDelay", self._input_delay_tracker.get()
            )
            if my_train_id == 0:
                my_train_id = self._last_train_id_processed + 1
                warn(
                    "Failed to get current train ID, using previously seen train "
                    "ID for future train thresholding - if this persists, check "
                    "connection to timeserver."
                )
        with self.warning_context(
            "inputDataState", WarningLampType.TRAIN_ID
        ) as warn:
            if train_id > (
                my_train_id
                + self.unsafe_get("workarounds.trainFromFutureThreshold")
            ):
                warn(
                    f"Suspecting train from the future: 'now' is {my_train_id}, "
                    f"received train ID {train_id}, dropping data"
                )
                raise TrainFromTheFutureException()
            try:
                self._train_ratio_tracker.update(train_id)
                self._buffered_status_update.set(
                    "performance.ratioOfRecentTrainsReceived",
                    self._train_ratio_tracker.get(
                        current_train=my_train_id,
                        expected_delay=math.ceil(
                            self.unsafe_get("performance.inputDelay") / 100
                        ),
                    )
                    if my_train_id != 0
                    else self._train_ratio_tracker.get(),
                )
            except trackers.NonMonotonicTrainIdWarning as ex:
                warn(
                    f"Train ratio tracker noticed issue with train ID: {ex}\n"
                    f"For the record, I think now is: {my_train_id}"
                )
                self._train_ratio_tracker.reset()
                self._train_ratio_tracker.update(train_id)

        self._last_train_id_processed = train_id
        self._buffered_status_update.set("trainId", train_id)

    def _update_rate_and_state(self):
        if self.get("state") is State.PROCESSING:
            # always update rate: estimate depends on query time
            self._buffered_status_update.set(
                "performance.rate", self._rate_tracker.get()
            )
            # remaining stats are set _buffered_status_update in input handler
            self.set(self._buffered_status_update)
            self._buffered_status_update.clear()
            if (
                default_timer() - self._last_processing_started
                > PROCESSING_STATE_TIMEOUT
            ):
                self.updateState(State.ON)
                self.log_status_info(
                    f"No new train in {PROCESSING_STATE_TIMEOUT} s, switching state."
                )
                self._train_ratio_tracker.reset()
                self._maybe_update_cell_and_pulse_tables(None, None)

    def _maybe_update_cell_and_pulse_tables(self, cell_table, pulse_table):
        if not np.array_equal(cell_table, self._last_cell_table):
            self._last_cell_table = cell_table
            if cell_table is None:
                self.set("dataFormat.cellId", [])
            else:
                self.set("dataFormat.cellId", list(map(int, cell_table)))
        if not np.array_equal(pulse_table, self._last_pulse_table):
            self._last_pulse_table = pulse_table
            if pulse_table is None:
                self.set("dataFormat.pulseId", [])
            else:
                self.set("dataFormat.pulseId", list(map(int, pulse_table)))

    def handle_eos(self, channel):
        self.updateState(State.ON)
        self.signalEndOfStream("dataOutput")


def add_addon_nodes(schema, device_class, prefix="addons"):
    det_name = device_class.__name__[:-len("Correction")].lower()
    device_class._available_addons = [
        addon.load()
        for addon in entry_points().select().get("calng.correction_addon", [])
        if not addon.extras
        or det_name in addon.extras
    ]

    for addon_class in device_class._available_addons:
        (
            NODE_ELEMENT(schema)
            .key(f"{prefix}.{addon_class.__name__}")
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{prefix}.{addon_class.__name__}.enable")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(False)
            .commit(),
        )
        addon_class.extend_device_schema(
            schema, f"{prefix}.{addon_class.__name__}"
        )
