import numpy as np
import pytest
from calng.corrections.LpdCorrection import (
    Constants,
    LpdBaseRunner,
    LpdCalcatFriend,
    LpdCpuRunner,
    LpdGpuRunner,
)

from common_setup import DummyCorrectionDevice, maybe_get


@pytest.fixture(params=[LpdCpuRunner, LpdGpuRunner])
def kernel_runner(request):
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return request.param(device)


@pytest.fixture
def cpu_runner():
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return LpdCpuRunner(device)


@pytest.fixture
def gpu_runner():
    device = DummyCorrectionDevice(LpdBaseRunner, LpdCalcatFriend)
    return LpdGpuRunner(device)


# generate some "raw data"
raw_data_shape = (100, 1, 256, 256)
# the image data part
raw_image = (np.arange(np.product(raw_data_shape)).reshape(raw_data_shape)).astype(
    np.uint16
)
# keeping in mind 12-bit ADC
raw_image = np.bitwise_and(raw_image, 0x0FFF)
raw_image_32 = raw_image.astype(np.float32)
# the gain values (TODO: test with some being illegal)
gain_data = (np.random.random(raw_data_shape) * 3).astype(np.uint16)
assert np.array_equal(gain_data, np.bitwise_and(gain_data, 0x0003))
gain_data_32 = gain_data.astype(np.float32)
raw_data = np.bitwise_or(raw_image, np.left_shift(gain_data, 12))
# just checking that we constructed something reasonable
assert np.array_equal(gain_data, np.bitwise_and(np.right_shift(raw_data, 12), 0x0003))
assert np.array_equal(raw_image, np.bitwise_and(raw_data, 0x0FFF))

# generate some defects
wrong_gain_value_indices = (np.random.random(raw_data_shape) < 0.005).astype(np.bool_)
raw_data_with_some_wrong_gain = raw_data.copy()
raw_data_with_some_wrong_gain[wrong_gain_value_indices] = np.bitwise_or(
    raw_data_with_some_wrong_gain[wrong_gain_value_indices], 0x3000
)

# generate some constants
dark_constant_shape = (256, 256, 512, 3)
# okay, as slow and fast scan have same size, can't tell the difference here...
funky_constant_shape = (256, 256, 512, 3)
offset_constant = np.random.random(dark_constant_shape).astype(np.float32)
gain_amp_constant = np.random.random(funky_constant_shape).astype(np.float32)
bp_dark_constant = (np.random.random(dark_constant_shape) < 0.01).astype(np.uint32)

cell_table = np.linspace(0, 512, 100).astype(np.uint16)
np.random.shuffle(cell_table)


def test_only_cast(kernel_runner):
    _, processed, previews = kernel_runner.correct(raw_data, cell_table)
    assert np.allclose(maybe_get(processed), raw_image_32[:, 0])
    # TODO: make raw preview show image data (remove gain bits)
    assert np.allclose(previews[0], raw_data[:, 0])  # raw: raw
    assert np.allclose(previews[1], raw_image_32[:, 0])  # processed: raw
    assert np.allclose(previews[2], gain_data_32[:, 0])  # gain: unchanged


def test_only_cast_with_some_wrong_gain(kernel_runner):
    (
        _,
        processed,
        _,
    ) = kernel_runner.correct(raw_data_with_some_wrong_gain, cell_table)
    assert np.all(np.isnan(processed[wrong_gain_value_indices[:, 0]]))


def test_correct(cpu_runner, gpu_runner):
    # naive numpy version way too slow, just compare the two runners
    # the CPU one uses kernel almost identical to pycalibration
    cpu_runner.load_constant(Constants.Offset, offset_constant)
    cpu_runner.load_constant(Constants.GainAmpMap, gain_amp_constant)
    cpu_runner.load_constant(Constants.BadPixelsDark, bp_dark_constant)
    gpu_runner.load_constant(Constants.Offset, offset_constant)
    gpu_runner.load_constant(Constants.GainAmpMap, gain_amp_constant)
    gpu_runner.load_constant(Constants.BadPixelsDark, bp_dark_constant)
    _, processed_cpu, _ = cpu_runner.correct(raw_data_with_some_wrong_gain, cell_table)
    _, processed_gpu, _ = gpu_runner.correct(raw_data_with_some_wrong_gain, cell_table)
    assert np.allclose(processed_cpu, processed_gpu.get(), equal_nan=True)
